package co.dapi.dapi_connect


import android.app.Application
import android.util.Log
import androidx.annotation.NonNull
import com.dapi.connect.core.base.DapiClient
import com.dapi.connect.core.callbacks.OnDapiConnectListener
import com.dapi.connect.core.callbacks.OnDapiTransferListener
import com.dapi.connect.core.enums.DapiEnvironment
import com.dapi.connect.core.enums.DapiTheme
import com.dapi.connect.data.endpoint_models.BeneficiaryCoolDownPeriod
import com.dapi.connect.data.models.*
import com.google.gson.Gson
import com.google.gson.reflect.TypeToken
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.EventChannel
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashMap


/** DapiConnectPlugin */
class DapiConnectPlugin : FlutterPlugin, MethodCallHandler, ActivityAware {
    /// The MethodChannel that will the communication between Flutter and native Android
    ///
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private lateinit var channel: MethodChannel
    private lateinit var application: Application

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, "dapi_connect")
        channel.setMethodCallHandler(this)
    }


    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        when (call.method) {
            "initClient:" -> {
                val configs = (call.arguments as List<*>).first() as Map<*, *>
                initClient(configs)
            }

            "presentConnect:" -> {
                presentConnect(result)
            }

            "presentAutoFlow:" -> {
                presentAutoFlow(result)
            }

            "dismissConnect:" -> {
                dismissConnect()
            }

            "dismissAutoFlow:" -> {
                dismissAutoFlow()
            }

            "setAutoFlowMinimumAmount:configurations:" -> {
                val amount = (call.arguments as List<*>).first() as Double
                setAutoFlowMinimumAmount(amount)

            }

            "setAutoFlowMaximumAmount:configurations:" -> {
                val amount = (call.arguments as List<*>).first() as Double
                setAutoFlowMaximumAmount(amount)
            }

            "setUserID:configurations:" -> {
                val userID = (call.arguments as List<*>).first() as String
                setUserID(userID)
            }

            "getUserID:" -> {
                getUserID(result)
            }

            "dapiConnections:" -> {
                dapiConnections(result)
            }

            "getIdentity:" -> {
                getIdentity(result)
            }

            "getAccounts:" -> {
                getAccounts(result)
            }

            "getBalance:configurations:" -> {
                val accountID = (call.arguments as List<*>).first() as String
                getBalance(accountID, result)
            }

            "getTransactions:fromTimestamp:toTimestamp:configurations:" -> {
                val params = (call.arguments as List<*>)
                val accountID = params.first() as String
                val fromDate = params[1] as Long
                val toDate = params[2] as Long
                getTransactions(accountID, Date(fromDate), Date(toDate), result)
            }

            "getAccountMetadata:" -> {
                getAccountMetaData(result)
            }

            "delinkUser:" -> {
                delinkUser(result)
            }

            "getBeneficiaries:" -> {
                getBeneficiaries(result)
            }

            "createBeneficiary:configurations:" -> {
                val beneficiaryInfoMap = (call.arguments as List<*>).first() as Map<*, *>
                val beneficiaryInfo = extractBeneficiaryInfo(beneficiaryInfoMap)
                createBeneficiary(beneficiaryInfo, result)
            }

            "createTransferWithReceiverID:" -> {
                val params = (call.arguments as List<*>)
                val receiverID = params.first() as String
                val senderID = params[1] as String
                val amount = params[2] as Double

                createTransfer(receiverID, senderID, amount, null, result)
            }

            "createTransferWithAccountNumber:" -> {
                val params = (call.arguments as List<*>)
                val accountNumber = params.first() as String
                val senderID = params[1] as String
                val amount = params[2] as Double
                val name = params[3] as String

                createTransfer(accountNumber, name, amount, senderID, null, result)
            }

            "createTransferWithIBAN:" -> {
                val params = (call.arguments as List<*>)
                val iban = params.first() as String
                val senderID = params[1] as String
                val amount = params[2] as Double
                val name = params[3] as String
                createTransfer(iban, name, senderID, amount, null, result)
            }

            else -> {
                result.notImplemented()
            }
        }
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
    }

    private fun initClient(configs: Map<*, *>) {
        val configurations = extractConfigurations(configs)
        val client = DapiClient(application, configurations)
    }

    private fun presentConnect(result: Result) {
        DapiClient.getInstance().connect.present()
        addConnectListener(result)
    }

    private fun addConnectListener(result: Result) {
        DapiClient.getInstance().connect.setOnConnectListener(object : OnDapiConnectListener {

            override fun onConnectionSuccessful(userID: String, bankID: String) {
                val resultMap = HashMap<String, String>()
                resultMap["bankID"] = bankID
                resultMap["userID"] = userID
                result.success(resultMap)
            }

            override fun onConnectionFailure(error: DapiError, bankID: String) {
                val resultMap = HashMap<String, Any>()
                resultMap["bankID"] = bankID
                resultMap["error"] = error
                result.error("1012", "Failed to connect to bank", resultMap)
            }

            override fun onProceed(userID: String, bankID: String) {

            }

            override fun setBeneficiaryInfoOnConnect(bankID: String): DapiBeneficiaryInfo? {
                return null
            }
        })
    }

    private fun presentAutoFlow(result: Result) {
        DapiClient.getInstance().autoFlow.present()
        addAutoFlowListener(result)
    }

    private fun addAutoFlowListener(result: Result){
        DapiClient.getInstance().autoFlow.setOnTransferListener(object : OnDapiTransferListener{
            override fun onAutoFlowSuccessful(amount: Double, senderAccountID: String?, recipientAccountID: String?, jobID: String) {
                val resultMap = HashMap<String, Any?>()
                resultMap["amount"] = amount
                resultMap["senderID"] = senderAccountID
                resultMap["receiverID"] = recipientAccountID
                resultMap["jobID"] = jobID
                result.success(resultMap)
            }

            override fun onAutoFlowFailure(error: DapiError, senderAccountID: String?, recipientAccountID: String?) {
                val resultMap = HashMap<String, Any?>()
                resultMap["senderID"] = senderAccountID
                resultMap["receiverID"] = recipientAccountID
                resultMap["error"] = error
                result.error("1012","Failed to send money",resultMap)
            }

            override fun onPaymentStarted(bankID: String) {

            }

            override fun setBeneficiaryInfoOnAutoFlow(bankID: String): DapiBeneficiaryInfo {
                return null!!
            }
        })
    }

    private fun dismissConnect() {
        DapiClient.getInstance().connect.dismiss()
    }

    private fun dismissAutoFlow() {
        DapiClient.getInstance().autoFlow.dismiss()
    }

    private fun setUserID(userID: String) {
        DapiClient.getInstance().userID = userID
    }

    private fun getUserID(result: Result) {
        val userID = DapiClient.getInstance().userID
        if (userID != null) {
            result.success(userID)
        } else {
            result.error("1012", "UserID is null", userID)
        }
    }

    private fun setAutoFlowMaximumAmount(amount: Double) {
        DapiClient.getInstance().autoFlow.maximumAmount = amount.toInt()
    }

    private fun setAutoFlowMinimumAmount(amount: Double) {
        DapiClient.getInstance().autoFlow.minimumAmount = amount.toInt()
    }

    private fun dapiConnections(result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.connect.getConnections({
            val resultConnections = mutableListOf<Map<String, Any>>()
            it.forEach { connection ->
                val resultAccountsMap = mutableListOf<Map<String, Any>>()
                connection.subAccounts.forEach { account ->
                    val currencyMap = hashMapOf<String, String>()
                    currencyMap["code"] = account.currency?.code!!
                    currencyMap["name"] = account.currency?.name!!

                    val accountMap = hashMapOf<String, Any>()
                    accountMap["iban"] = account.iban!!
                    accountMap["number"] = account.number!!
                    accountMap["currency"] = currencyMap
                    accountMap["type"] = account.type!!
                    accountMap["name"] = account.name!!
                    accountMap["id"] = account.id!!
                    accountMap["isFavourite"] = account.isFavourite
                    resultAccountsMap.add(accountMap)
                }

                val connectionMap = hashMapOf<String, Any>()
                connectionMap["userID"] = connection.userID
                connectionMap["clientUserID"] = connection.clientUserID
                connectionMap["bankID"] = connection.bankID
                connectionMap["bankName"] = connection.shortBankName
                connectionMap["beneficiaryCoolDownPeriod"] = convertCoolDownPeriodToMillis(connection.coolDownPeriod).toDouble()
                connectionMap["countryName"] = connection.country
                connectionMap["isCreateBeneficiaryEndpointRequired"] = connection.isCreateBeneficiaryRequired
                connectionMap["accounts"] = resultAccountsMap
                resultConnections.add(connectionMap)
            }

            result.success(resultConnections)
        }, {
            result.error("1012", it?.msg, it.serializeToMap())
        })
    }

    private fun getIdentity(result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.data.getIdentity({
            val identity = it.identity.serializeToMap()
            result.success(identity)
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun getAccounts(result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.data.getAccounts({
            val accounts = it.accounts.serializeToMap()
            result.success(accounts)
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun getBalance(accountID: String, result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.data.getBalance(accountID, {
            val balance = it.balance.serializeToMap()
            result.success(balance)
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun getTransactions(accountID: String, fromDate: Date, toDate: Date, result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.data.getTransactions(accountID, fromDate, toDate, {
            val transactions = it.transactions.serializeToMap()
            result.success(transactions)
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun getAccountMetaData(result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.metadata.getAccountMetaData({
            val accountMetaData = it.accountsMetadata.serializeToMap()
            result.success(accountMetaData)
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun delinkUser(result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.auth.delink({
            result.success(it.serializeToMap())
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun getBeneficiaries(result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.payment.getBeneficiaries({
            val beneficiaries = it.beneficiaries.serializeToMap()
            result.success(beneficiaries)
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun createBeneficiary(beneficiaryInfo: DapiBeneficiaryInfo, result: Result) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.payment.createBeneficiary(beneficiaryInfo, {
            result.success(it.serializeToMap())
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun createTransfer(
            receiverID: String,
            senderID: String,
            amount: Double,
            remark: String?,
            result: Result
    ) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.payment.createTransfer(receiverID, senderID, amount, remark, {
            result.success(it.serializeToMap())
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun createTransfer(
            iban: String,
            name: String,
            senderID: String,
            amount: Double,
            remark: String?,
            result: Result
    ) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.payment.createTransfer(iban, name, senderID, amount, remark, {
            result.success(it.serializeToMap())
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private fun createTransfer(
            accountNumber: String,
            name: String,
            amount: Double,
            senderID: String,
            remark: String?,
            result: Result
    ) {
        val dapiClient = DapiClient.getInstance()
        dapiClient.payment.createTransfer(accountNumber, name, amount, senderID, remark, {
            result.success(it.serializeToMap())
        }, {
            result.error("1012", it.msg, it.serializeToMap())
        })
    }

    private val gson = Gson()

    private fun <T> T.serializeToMap(): Map<String, Any?> {
        return convert()
    }

    private inline fun <reified T> Map<String, Any?>.toDataClass(): T {
        return convert()
    }

    private inline fun <I, reified O> I.convert(): O {
        val json = gson.toJson(this)
        return gson.fromJson(json, object : TypeToken<O>() {}.type)
    }

    private fun extractBeneficiaryInfo(beneficiaryInfoMap: Map<*, *>): DapiBeneficiaryInfo {
        val linesMap = beneficiaryInfoMap["address"] as Map<String, String>
        val line1 = linesMap["line1"]
        val line2 = linesMap["line2"]
        val line3 = linesMap["line3"]
        val linesAddress = LinesAddress(
                line1,
                line2,
                line3
        )
        val accountNumber = beneficiaryInfoMap["accountNumber"] as String
        val name = beneficiaryInfoMap["name"] as String
        val bankName = beneficiaryInfoMap["bankName"] as String
        val swiftCode = beneficiaryInfoMap["swiftCode"] as String
        val iban = beneficiaryInfoMap["iban"] as String
        val phoneNumber = beneficiaryInfoMap["phoneNumber"] as String
        val country = beneficiaryInfoMap["country"] as String
        val branchAddress = beneficiaryInfoMap["branchAddress"] as String
        val branchName = beneficiaryInfoMap["branchName"] as String

        return DapiBeneficiaryInfo(
                linesAddress,
                accountNumber,
                name,
                bankName,
                swiftCode,
                iban,
                country,
                branchAddress,
                branchName,
                phoneNumber
        )
    }


    private fun extractConfigurations(configurations: Map<*, *>): DapiConfigurations {
        var appKey: String? = null
        var clientUserID: String? = null
        var environment: String? = null
        var baseUrl: String? = null
        var countries: ArrayList<Any>? = null
        val endpoints: DapiEndpoints?


        if (!configurations.containsKey("appKey")) {
            Log.e("DapiError", "Invalid appKey argument")
        } else {
            appKey = configurations["appKey"] as String
        }

        if (!configurations.containsKey("clientUserID")) {
            Log.e("DapiError", "Invalid clientUserID argument")
        } else {
            clientUserID = configurations["clientUserID"] as String
        }

        val userID: String? = if (!configurations.containsKey("userID")) {
            null
        } else {
            configurations["userID"] as String
        }

        val colorScheme: String? = if (!configurations.containsKey("colorScheme")) {
            "GENERAL"
        } else {
            configurations["colorScheme"] as String
        }

        if (!configurations.containsKey("environment")) {
            Log.e("DapiError", "Invalid environment argument")
        } else {
            environment = configurations["environment"] as String
        }

        if (!configurations.containsKey("countries")) {
            Log.e("DapiError", "Invalid countries argument")
        } else {
            countries = configurations["countries"] as ArrayList<Any>
        }

        val isExperimental: Boolean = if (!configurations.containsKey("isExperimental")) {
            false
        } else {
            configurations["isExperimental"] as Boolean
        }

        if (!configurations.containsKey("baseURL")) {
            Log.e("DapiError", "Invalid baseURL argument")
        } else {
            baseUrl = configurations["baseURL"] as String
        }

        val extraParams: Map<String, Any>? = if (!configurations.containsKey("endpointExtraQueryItems")) {
            HashMap()
        } else {
            configurations["endpointExtraQueryItems"] as HashMap<String, Any>
        }

        val extraHeaders: Map<String, Any>? = if (!configurations.containsKey("endpointExtraHeaderFields")) {
            HashMap()
        } else {
            configurations["endpointExtraHeaderFields"] as HashMap<String, Any>
        }

        val extraBody: Map<String, Any>? = if (!configurations.containsKey("endpointExtraBody")) {
            HashMap()
        } else {
            configurations["endpointExtraBody"] as HashMap<String, Any>
        }

        if (!configurations.containsKey("endpoints")) {
            endpoints = DapiEndpoints()
        } else {
            val endpointsMap: Map<String, Any?> = configurations["endpoints"] as HashMap<String, Any>
            val getIdentity: String?
            val getAccounts: String?
            val getBalance: String?
            val getTransactions: String?
            val accountMetaData: String?
            val createTransfer: String?
            val createBeneficiary: String?
            val getBeneficiaries: String?
            val delinkUser: String?
            val resumeJob: String?
            getIdentity = if (endpointsMap["getIdentity"] == null) {
                DapiEndpoints.GET_IDENTITY_ENDPOINT
            } else {
                endpointsMap["getIdentity"] as String?
            }
            getAccounts = if (endpointsMap["getAccounts"] == null) {
                DapiEndpoints.GET_ACCOUNTS_ENDPOINT
            } else {
                endpointsMap["getAccounts"] as String?
            }
            getBalance = if (endpointsMap["getBalance"] == null) {
                DapiEndpoints.GET_BALANCE_ENDPOINT
            } else {
                endpointsMap["getBalance"] as String?
            }
            getTransactions = if (endpointsMap["getTransactions"] == null) {
                DapiEndpoints.GET_TRANSACTIONS_ENDPOINT
            } else {
                endpointsMap["getTransactions"] as String?
            }
            accountMetaData = if (endpointsMap["getAccountMetadata"] == null) {
                DapiEndpoints.ACCOUNT_META_DATA_ENDPOINT
            } else {
                endpointsMap["getAccountMetadata"] as String?
            }
            createTransfer = if (endpointsMap["createTransfer"] == null) {
                DapiEndpoints.CREATE_TRANSFER_ENDPOINT
            } else {
                endpointsMap["createTransfer"] as String?
            }
            createBeneficiary = if (endpointsMap["createBeneficiary"] == null) {
                DapiEndpoints.CREATE_BENEFICIARY_ENDPOINT
            } else {
                endpointsMap["createBeneficiary"] as String?
            }
            getBeneficiaries = if (endpointsMap["getBeneficiaries"] == null) {
                DapiEndpoints.GET_BENEFICIARIES_ENDPOINT
            } else {
                endpointsMap["getBeneficiaries"] as String?
            }
            delinkUser = if (endpointsMap["delinkUser"] == null) {
                DapiEndpoints.DELINK_USER_ENDPOINT
            } else {
                endpointsMap["delinkUser"] as String?
            }
            resumeJob = if (endpointsMap["resumeJob"] == null) {
                DapiEndpoints.RESUME_JOB_ENDPOINT
            } else {
                endpointsMap["resumeJob"] as String?
            }
            endpoints = DapiEndpoints(
                    getIdentity!!,
                    getAccounts!!,
                    getBalance!!,
                    getTransactions!!,
                    accountMetaData!!,
                    createTransfer!!,
                    createBeneficiary!!,
                    getBeneficiaries!!,
                    delinkUser!!,
                    resumeJob!!
            )
        }

        val dapiEnvironment = if (environment!!.toLowerCase().equals("production")) {
            DapiEnvironment.PRODUCTION;
        } else {
            DapiEnvironment.SANDBOX;
        }

        val theme = if (colorScheme?.toLowerCase().equals("neon")) {
            DapiTheme.ELECTRIC;
        } else if (colorScheme?.toLowerCase().equals("blackandwhite")) {
            DapiTheme.ELEGANT;
        } else {
            DapiTheme.GENERAL;
        }

        val extraBodyMap = if (extraBody is HashMap) extraBody else HashMap(extraBody)

        val extraHeadersMap = if (extraHeaders is HashMap) extraHeaders else HashMap(extraHeaders)

        val extraParamsMap = if (extraParams is HashMap) extraParams else HashMap(extraParams)


        val finalExtraBodyMap: HashMap<String, String> = HashMap()
        for ((key, value) in extraBodyMap) {
            if (value is String) {
                finalExtraBodyMap[key] = value as String
            }
        }

        val finalExtraHeadersMap: HashMap<String, String> = HashMap()
        for ((key, value) in extraHeadersMap) {
            if (value is String) {
                finalExtraHeadersMap[key] = value as String
            }
        }

        val finalExtraParamsMap: HashMap<String, String> = HashMap()
        for ((key, value) in extraParamsMap) {
            if (value is String) {
                finalExtraParamsMap[key] = value as String
            }
        }

        val finalCountries = ArrayList<String>(countries!!.size)
        for (`object` in countries) {
            finalCountries.add(`object`.toString())
        }

        return DapiConfigurations(
                appKey!!,
                baseUrl!!,
                dapiEnvironment,
                finalCountries,
                clientUserID,
                userID,
                isExperimental,
                theme,
                finalExtraHeadersMap,
                finalExtraParamsMap,
                finalExtraBodyMap,
                endpoints
        )


    }

    private fun convertCoolDownPeriodToMillis(coolDownPeriod: BeneficiaryCoolDownPeriod): Long {
        val value = coolDownPeriod.value

        return when (coolDownPeriod.unit) {
            "hrs" -> {
                value * 3600000L
            }
            "days" -> {
                value * 86400000L
            }
            else -> {
                0L
            }
        }
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        application = binding.activity.application
    }

    override fun onDetachedFromActivityForConfigChanges() {
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        application = binding.activity.application
    }

    override fun onDetachedFromActivity() {
    }
}
